import { Configuration } from './utils';
import { EntityFactory, EntityRepository, EntityValidator } from './entity';
import { LockMode, UnitOfWork } from './unit-of-work';
import { FilterQuery, IDatabaseDriver } from './drivers';
import { EntityData, EntityName, IEntity, IEntityType, IPrimaryKey } from './decorators';
import { QueryBuilder, QueryOrderMap } from './query';
import { Connection } from './connections';
export declare class EntityManager {
    readonly config: Configuration;
    private readonly driver;
    private readonly validator;
    private readonly repositoryMap;
    private readonly entityLoader;
    private readonly metadata;
    private readonly unitOfWork;
    private readonly entityFactory;
    constructor(config: Configuration, driver: IDatabaseDriver);
    getDriver<D extends IDatabaseDriver = IDatabaseDriver>(): D;
    getConnection<C extends Connection = Connection>(): C;
    getRepository<T extends IEntityType<T>>(entityName: EntityName<T>): EntityRepository<T>;
    getValidator(): EntityValidator;
    createQueryBuilder(entityName: EntityName<IEntity>, alias?: string): QueryBuilder;
    find<T extends IEntityType<T>>(entityName: EntityName<T>, where?: FilterQuery<T>, options?: FindOptions): Promise<T[]>;
    find<T extends IEntityType<T>>(entityName: EntityName<T>, where?: FilterQuery<T>, populate?: string[], orderBy?: QueryOrderMap, limit?: number, offset?: number): Promise<T[]>;
    findOne<T extends IEntityType<T>>(entityName: EntityName<T>, where: FilterQuery<T> | IPrimaryKey, options?: FindOneOptions): Promise<T | null>;
    findOne<T extends IEntityType<T>>(entityName: EntityName<T>, where: FilterQuery<T> | IPrimaryKey, populate?: string[], orderBy?: QueryOrderMap): Promise<T | null>;
    beginTransaction(): Promise<void>;
    commit(): Promise<void>;
    rollback(): Promise<void>;
    transactional(cb: (em: EntityManager) => Promise<any>): Promise<any>;
    lock(entity: IEntity, lockMode: LockMode, lockVersion?: number | Date): Promise<void>;
    nativeInsert<T extends IEntityType<T>>(entityName: EntityName<T>, data: EntityData<T>): Promise<IPrimaryKey>;
    nativeUpdate<T extends IEntityType<T>>(entityName: EntityName<T>, where: FilterQuery<T>, data: EntityData<T>): Promise<number>;
    nativeDelete<T extends IEntityType<T>>(entityName: EntityName<T>, where: FilterQuery<T> | string | any): Promise<number>;
    map<T extends IEntityType<T>>(entityName: EntityName<T>, result: EntityData<T>): T;
    /**
     * Shortcut to driver's aggregate method. Available in MongoDriver only.
     */
    aggregate(entityName: EntityName<IEntity>, pipeline: any[]): Promise<any[]>;
    merge<T extends IEntityType<T>>(entity: T, refresh?: boolean): T;
    merge<T extends IEntityType<T>>(entityName: EntityName<T>, data: EntityData<T>, refresh?: boolean): T;
    /**
     * Creates new instance of given entity and populates it with given data
     */
    create<T extends IEntityType<T>>(entityName: EntityName<T>, data: EntityData<T>): T;
    /**
     * Gets a reference to the entity identified by the given type and identifier without actually loading it, if the entity is not yet loaded
     */
    getReference<T extends IEntityType<T>>(entityName: EntityName<T>, id: IPrimaryKey): T;
    count<T extends IEntityType<T>>(entityName: EntityName<T>, where: FilterQuery<T>): Promise<number>;
    persist(entity: IEntity | IEntity[], flush?: boolean): Promise<void>;
    persistAndFlush(entity: IEntity | IEntity[]): Promise<void>;
    persistLater(entity: IEntity | IEntity[]): void;
    remove<T extends IEntityType<T>>(entityName: EntityName<T>, where: T | any, flush?: boolean): Promise<number>;
    removeEntity(entity: IEntity, flush?: boolean): Promise<void>;
    removeAndFlush(entity: IEntity): Promise<void>;
    removeLater(entity: IEntity): void;
    /**
     * flush changes to database
     */
    flush(): Promise<void>;
    /**
     * clear identity map, detaching all entities
     */
    clear(): void;
    canPopulate(entityName: string | Function, property: string): boolean;
    fork(clear?: boolean): EntityManager;
    getUnitOfWork(): UnitOfWork;
    getEntityFactory(): EntityFactory;
    private checkLockRequirements;
    private lockAndPopulate;
}
export interface FindOptions {
    populate?: string[];
    orderBy?: QueryOrderMap;
    limit?: number;
    offset?: number;
}
export interface FindOneOptions {
    populate?: string[];
    orderBy?: QueryOrderMap;
    lockMode?: LockMode;
    lockVersion?: number | Date;
    refresh?: boolean;
    fields?: string[];
}
