"use strict";
Object.defineProperty(exports, "__esModule", { value: true });
const utils_1 = require("./utils");
const entity_1 = require("./entity");
const unit_of_work_1 = require("./unit-of-work");
const query_1 = require("./query");
const metadata_1 = require("./metadata");
class EntityManager {
    constructor(config, driver) {
        this.config = config;
        this.driver = driver;
        this.validator = new entity_1.EntityValidator(this.config.get('strict'));
        this.repositoryMap = {};
        this.entityLoader = new entity_1.EntityLoader(this);
        this.metadata = metadata_1.MetadataStorage.getMetadata();
        this.unitOfWork = new unit_of_work_1.UnitOfWork(this);
        this.entityFactory = new entity_1.EntityFactory(this.unitOfWork, this.driver, this.config);
    }
    getDriver() {
        return this.driver;
    }
    getConnection() {
        return this.driver.getConnection();
    }
    getRepository(entityName) {
        entityName = utils_1.Utils.className(entityName);
        if (!this.repositoryMap[entityName]) {
            const meta = this.metadata[entityName];
            const RepositoryClass = this.config.getRepositoryClass(meta.customRepository);
            this.repositoryMap[entityName] = new RepositoryClass(this, entityName);
        }
        return this.repositoryMap[entityName];
    }
    getValidator() {
        return this.validator;
    }
    createQueryBuilder(entityName, alias) {
        entityName = utils_1.Utils.className(entityName);
        return new query_1.QueryBuilder(entityName, this.metadata, this.driver, alias);
    }
    async find(entityName, where = {}, populate, orderBy, limit, offset) {
        entityName = utils_1.Utils.className(entityName);
        where = query_1.SmartQueryHelper.processWhere(where, entityName);
        this.validator.validateParams(where);
        const options = utils_1.Utils.isObject(populate) ? populate : { populate, orderBy, limit, offset };
        const results = await this.driver.find(entityName, where, options.populate || [], options.orderBy || {}, options.limit, options.offset);
        if (results.length === 0) {
            return [];
        }
        const ret = [];
        for (const data of results) {
            const entity = this.merge(entityName, data);
            ret.push(entity);
        }
        await this.entityLoader.populate(entityName, ret, options.populate || []);
        return ret;
    }
    async findOne(entityName, where, populate, orderBy) {
        entityName = utils_1.Utils.className(entityName);
        const options = utils_1.Utils.isObject(populate) ? populate : { populate, orderBy };
        this.validator.validateEmptyWhere(where);
        where = query_1.SmartQueryHelper.processWhere(where, entityName);
        this.checkLockRequirements(options.lockMode, this.metadata[entityName]);
        let entity = this.getUnitOfWork().tryGetById(entityName, where);
        const isOptimisticLocking = !utils_1.Utils.isDefined(options.lockMode) || options.lockMode === unit_of_work_1.LockMode.OPTIMISTIC;
        if (entity && entity.isInitialized() && !options.refresh && isOptimisticLocking) {
            return this.lockAndPopulate(entity, options);
        }
        this.validator.validateParams(where);
        const data = await this.driver.findOne(entityName, where, options.populate, options.orderBy, options.fields, options.lockMode);
        if (!data) {
            return null;
        }
        entity = this.merge(entityName, data, options.refresh);
        await this.lockAndPopulate(entity, options);
        return entity;
    }
    async beginTransaction() {
        await this.driver.beginTransaction();
    }
    async commit() {
        await this.driver.commit();
    }
    async rollback() {
        await this.driver.rollback();
    }
    async transactional(cb) {
        const em = this.fork(false);
        await em.getDriver().transactional(async () => {
            const ret = await cb(em);
            await em.flush();
            return ret;
        });
    }
    async lock(entity, lockMode, lockVersion) {
        await this.getUnitOfWork().lock(entity, lockMode, lockVersion);
    }
    async nativeInsert(entityName, data) {
        entityName = utils_1.Utils.className(entityName);
        data = query_1.SmartQueryHelper.processParams(data);
        this.validator.validateParams(data, 'insert data');
        const res = await this.driver.nativeInsert(entityName, data);
        return res.insertId;
    }
    async nativeUpdate(entityName, where, data) {
        entityName = utils_1.Utils.className(entityName);
        data = query_1.SmartQueryHelper.processParams(data);
        where = query_1.SmartQueryHelper.processWhere(where, entityName);
        this.validator.validateParams(data, 'update data');
        this.validator.validateParams(where, 'update condition');
        const res = await this.driver.nativeUpdate(entityName, where, data);
        return res.affectedRows;
    }
    async nativeDelete(entityName, where) {
        entityName = utils_1.Utils.className(entityName);
        where = query_1.SmartQueryHelper.processWhere(where, entityName);
        this.validator.validateParams(where, 'delete condition');
        const res = await this.driver.nativeDelete(entityName, where);
        return res.affectedRows;
    }
    map(entityName, result) {
        entityName = utils_1.Utils.className(entityName);
        const meta = this.metadata[entityName];
        const data = this.driver.mapResult(result, meta);
        return this.merge(entityName, data, true);
    }
    /**
     * Shortcut to driver's aggregate method. Available in MongoDriver only.
     */
    async aggregate(entityName, pipeline) {
        entityName = utils_1.Utils.className(entityName);
        return this.driver.aggregate(entityName, pipeline);
    }
    merge(entityName, data, refresh) {
        if (utils_1.Utils.isEntity(entityName)) {
            return this.merge(entityName.constructor.name, entityName, data);
        }
        entityName = utils_1.Utils.className(entityName);
        this.validator.validatePrimaryKey(data, this.metadata[entityName]);
        let entity = this.getUnitOfWork().tryGetById(entityName, data);
        if (entity && entity.isInitialized() && !refresh) {
            return entity;
        }
        entity = utils_1.Utils.isEntity(data) ? data : this.getEntityFactory().create(entityName, data, true);
        // add to IM immediately - needed for self-references that can be part of `data` (and do not trigger cascade merge)
        this.getUnitOfWork().merge(entity, [entity]);
        entity_1.EntityAssigner.assign(entity, data, true);
        this.getUnitOfWork().merge(entity); // add to IM again so we have correct payload saved to change set computation
        return entity;
    }
    /**
     * Creates new instance of given entity and populates it with given data
     */
    create(entityName, data) {
        return this.getEntityFactory().create(entityName, data, false);
    }
    /**
     * Gets a reference to the entity identified by the given type and identifier without actually loading it, if the entity is not yet loaded
     */
    getReference(entityName, id) {
        const entity = this.getEntityFactory().createReference(entityName, id);
        this.getUnitOfWork().merge(entity, [], false);
        return entity;
    }
    async count(entityName, where) {
        entityName = utils_1.Utils.className(entityName);
        where = query_1.SmartQueryHelper.processWhere(where, entityName);
        this.validator.validateParams(where);
        return this.driver.count(entityName, where);
    }
    async persist(entity, flush = this.config.get('autoFlush')) {
        if (flush) {
            await this.persistAndFlush(entity);
        }
        else {
            this.persistLater(entity);
        }
    }
    async persistAndFlush(entity) {
        entity = utils_1.Utils.asArray(entity);
        for (const ent of entity) {
            this.getUnitOfWork().persist(ent);
        }
        await this.flush();
    }
    persistLater(entity) {
        entity = utils_1.Utils.asArray(entity);
        for (const ent of entity) {
            this.getUnitOfWork().persist(ent);
        }
    }
    async remove(entityName, where, flush = this.config.get('autoFlush')) {
        entityName = utils_1.Utils.className(entityName);
        if (utils_1.Utils.isEntity(where)) {
            await this.removeEntity(where, flush);
            return 1;
        }
        return this.nativeDelete(entityName, where);
    }
    async removeEntity(entity, flush = this.config.get('autoFlush')) {
        if (flush) {
            await this.removeAndFlush(entity);
        }
        else {
            this.removeLater(entity);
        }
    }
    async removeAndFlush(entity) {
        this.getUnitOfWork().remove(entity);
        await this.flush();
    }
    removeLater(entity) {
        this.getUnitOfWork().remove(entity);
    }
    /**
     * flush changes to database
     */
    async flush() {
        await this.getUnitOfWork().commit();
    }
    /**
     * clear identity map, detaching all entities
     */
    clear() {
        this.getUnitOfWork().clear();
    }
    canPopulate(entityName, property) {
        entityName = utils_1.Utils.className(entityName);
        const [p, ...parts] = property.split('.');
        const props = this.metadata[entityName].properties;
        const ret = p in props && props[p].reference !== entity_1.ReferenceType.SCALAR;
        if (!ret) {
            return false;
        }
        if (parts.length > 0) {
            return this.canPopulate(props[p].type, parts.join('.'));
        }
        return ret;
    }
    fork(clear = true) {
        const em = new EntityManager(this.config, this.driver);
        if (!clear) {
            Object.values(this.getUnitOfWork().getIdentityMap()).forEach(entity => em.merge(entity));
        }
        return em;
    }
    getUnitOfWork() {
        const em = utils_1.RequestContext.getEntityManager() || this;
        return em.unitOfWork;
    }
    getEntityFactory() {
        const em = utils_1.RequestContext.getEntityManager() || this;
        return em.entityFactory;
    }
    checkLockRequirements(mode, meta) {
        if (!mode) {
            return;
        }
        if (mode === unit_of_work_1.LockMode.OPTIMISTIC && !meta.versionProperty) {
            throw utils_1.ValidationError.notVersioned(meta);
        }
        if ([unit_of_work_1.LockMode.PESSIMISTIC_READ, unit_of_work_1.LockMode.PESSIMISTIC_WRITE].includes(mode) && !this.getDriver().isInTransaction()) {
            throw utils_1.ValidationError.transactionRequired();
        }
    }
    async lockAndPopulate(entity, options) {
        if (options.lockMode === unit_of_work_1.LockMode.OPTIMISTIC) {
            await this.lock(entity, options.lockMode, options.lockVersion);
        }
        await this.entityLoader.populate(entity.constructor.name, [entity], options.populate || []);
        return entity;
    }
}
exports.EntityManager = EntityManager;
