"use strict";
Object.defineProperty(exports, "__esModule", { value: true });
const mongodb_1 = require("mongodb");
const Connection_1 = require("./Connection");
const utils_1 = require("../utils");
const query_1 = require("../query");
class MongoConnection extends Connection_1.Connection {
    async connect() {
        this.client = await mongodb_1.MongoClient.connect(this.config.getClientUrl(), this.getConnectionOptions());
        this.db = this.client.db(this.config.get('dbName'));
    }
    async close(force) {
        return this.client.close(force);
    }
    async isConnected() {
        return this.client.isConnected();
    }
    getCollection(name) {
        return this.db.collection(this.getCollectionName(name));
    }
    getDefaultClientUrl() {
        return 'mongodb://127.0.0.1:27017';
    }
    getConnectionOptions() {
        const ret = { useNewUrlParser: true };
        const user = this.config.get('user');
        const password = this.config.get('password');
        if (user && password) {
            ret.auth = { user, password };
        }
        return ret;
    }
    getClientUrl() {
        const options = this.getConnectionOptions();
        const clientUrl = this.config.getClientUrl(true);
        const match = clientUrl.match(/^(\w+):\/\/((.*@.+)|.+)$/);
        return match ? `${match[1]}://${options.auth ? options.auth.user + ':*****@' : ''}${match[2]}` : clientUrl;
    }
    async execute(query) {
        throw new Error(`${this.constructor.name} does not support generic execute method`);
    }
    async find(collection, where, orderBy, limit, offset, fields) {
        collection = this.getCollectionName(collection);
        where = this.convertObjectIds(where);
        const options = {};
        if (fields) {
            options.projection = fields.reduce((o, k) => (Object.assign({}, o, { [k]: 1 })), {});
        }
        const resultSet = this.getCollection(collection).find(where, options);
        let query = `db.getCollection("${collection}").find(${JSON.stringify(where)}, ${JSON.stringify(options)})`;
        if (orderBy && Object.keys(orderBy).length > 0) {
            orderBy = Object.keys(orderBy).reduce((p, c) => {
                const direction = orderBy[c];
                return Object.assign({}, p, { [c]: utils_1.Utils.isString(direction) ? direction.toUpperCase() === query_1.QueryOrder.ASC ? 1 : -1 : direction });
            }, {});
            query += `.sort(${JSON.stringify(orderBy)})`;
            resultSet.sort(orderBy);
        }
        if (limit !== undefined) {
            query += `.limit(${limit})`;
            resultSet.limit(limit);
        }
        if (offset !== undefined) {
            query += `.skip(${offset})`;
            resultSet.skip(offset);
        }
        const now = Date.now();
        const res = await resultSet.toArray();
        this.logQuery(`${query}.toArray();`, Date.now() - now);
        return res;
    }
    async insertOne(collection, data) {
        collection = this.getCollectionName(collection);
        data = this.convertObjectIds(data);
        const now = Date.now();
        const res = await this.getCollection(collection).insertOne(data);
        const query = `db.getCollection("${collection}").insertOne(${JSON.stringify(data)});`;
        this.logQuery(query, Date.now() - now);
        return this.transformResult(res);
    }
    async updateMany(collection, where, data) {
        collection = this.getCollectionName(collection);
        where = this.convertObjectIds(where);
        data = this.convertObjectIds(data);
        const payload = Object.keys(data).some(k => k.startsWith('$')) ? data : { $set: data };
        const query = `db.getCollection("${collection}").updateMany(${JSON.stringify(where)}, ${JSON.stringify(payload)});`;
        const now = Date.now();
        const res = await this.getCollection(collection).updateMany(where, payload);
        this.logQuery(query, Date.now() - now);
        return this.transformResult(res);
    }
    async deleteMany(collection, where) {
        collection = this.getCollectionName(collection);
        where = this.convertObjectIds(where);
        const query = `db.getCollection("${collection}").deleteMany(${JSON.stringify(where)})`;
        const now = Date.now();
        const res = await this.getCollection(collection).deleteMany(where);
        this.logQuery(query, Date.now() - now);
        return this.transformResult(res);
    }
    async aggregate(collection, pipeline) {
        collection = this.getCollectionName(collection);
        const query = `db.getCollection("${collection}").aggregate(${JSON.stringify(pipeline)}).toArray();`;
        const now = Date.now();
        const res = this.getCollection(collection).aggregate(pipeline).toArray();
        this.logQuery(query, Date.now() - now);
        return res;
    }
    async countDocuments(collection, where) {
        collection = this.getCollectionName(collection);
        where = this.convertObjectIds(where);
        const query = `db.getCollection("${collection}").countDocuments(${JSON.stringify(where)})`;
        const now = Date.now();
        const res = await this.getCollection(collection).countDocuments(where);
        this.logQuery(query, Date.now() - now);
        return res;
    }
    convertObjectIds(payload) {
        if (payload instanceof mongodb_1.ObjectID) {
            return payload;
        }
        if (utils_1.Utils.isString(payload) && payload.match(/^[0-9a-f]{24}$/i)) {
            return new mongodb_1.ObjectID(payload);
        }
        if (Array.isArray(payload)) {
            return payload.map((item) => this.convertObjectIds(item));
        }
        if (utils_1.Utils.isObject(payload)) {
            Object.keys(payload).forEach(k => {
                payload[k] = this.convertObjectIds(payload[k]);
            });
        }
        return payload;
    }
    transformResult(res) {
        return {
            affectedRows: res.modifiedCount || res.deletedCount || 0,
            insertId: res.insertedId,
        };
    }
    getCollectionName(name) {
        name = utils_1.Utils.className(name);
        return this.metadata[name] ? this.metadata[name].collection : name;
    }
}
exports.MongoConnection = MongoConnection;
