"use strict";
Object.defineProperty(exports, "__esModule", { value: true });
const DatabaseDriver_1 = require("./DatabaseDriver");
const entity_1 = require("../entity");
const query_1 = require("../query");
const utils_1 = require("../utils");
class AbstractSqlDriver extends DatabaseDriver_1.DatabaseDriver {
    async find(entityName, where, populate = [], orderBy = {}, limit, offset) {
        const qb = this.createQueryBuilder(entityName);
        qb.select('*').populate(populate).where(where).orderBy(orderBy);
        if (limit !== undefined) {
            qb.limit(limit, offset);
        }
        return qb.execute('all');
    }
    async findOne(entityName, where, populate = [], orderBy = {}, fields, lockMode) {
        const pk = this.metadata[entityName].primaryKey;
        if (utils_1.Utils.isPrimaryKey(where)) {
            where = { [pk]: where };
        }
        if (fields && !fields.includes(pk)) {
            fields.unshift(pk);
        }
        return this.createQueryBuilder(entityName)
            .select(fields || '*')
            .populate(populate)
            .where(where)
            .orderBy(orderBy)
            .limit(1)
            .setLockMode(lockMode)
            .execute('get');
    }
    async count(entityName, where) {
        const qb = this.createQueryBuilder(entityName);
        const pk = this.metadata[entityName].primaryKey;
        const res = await qb.count(pk, true).where(where).execute('get', false);
        return +res.count;
    }
    async nativeInsert(entityName, data) {
        const collections = this.extractManyToMany(entityName, data);
        const pk = this.getPrimaryKeyField(entityName);
        if (Object.keys(data).length === 0) {
            data[pk] = null;
        }
        const qb = this.createQueryBuilder(entityName);
        const res = await qb.insert(data).execute('run', false);
        res.insertId = res.insertId || data[pk];
        await this.processManyToMany(entityName, res.insertId, collections);
        return res;
    }
    async nativeUpdate(entityName, where, data) {
        const pk = this.metadata[entityName] ? this.metadata[entityName].primaryKey : this.config.getNamingStrategy().referenceColumnName();
        if (utils_1.Utils.isPrimaryKey(where)) {
            where = { [pk]: where };
        }
        const collections = this.extractManyToMany(entityName, data);
        let res = { affectedRows: 0, insertId: 0 };
        if (Object.keys(data).length) {
            const qb = this.createQueryBuilder(entityName);
            res = await qb.update(data).where(where).execute('run', false);
        }
        await this.processManyToMany(entityName, utils_1.Utils.extractPK(data[pk] || where, this.metadata[entityName]), collections);
        return res;
    }
    async nativeDelete(entityName, where) {
        if (utils_1.Utils.isPrimaryKey(where)) {
            const pk = this.metadata[entityName] ? this.metadata[entityName].primaryKey : this.config.getNamingStrategy().referenceColumnName();
            where = { [pk]: where };
        }
        return this.createQueryBuilder(entityName).delete(where).execute('run', false);
    }
    createQueryBuilder(entityName) {
        return new query_1.QueryBuilder(entityName, this.metadata, this);
    }
    extractManyToMany(entityName, data) {
        if (!this.metadata[entityName]) {
            return {};
        }
        const props = this.metadata[entityName].properties;
        const ret = {};
        for (const k of Object.keys(data)) {
            const prop = props[k];
            if (prop && prop.reference === entity_1.ReferenceType.MANY_TO_MANY) {
                ret[k] = data[k];
                delete data[k];
            }
        }
        return ret;
    }
    async processManyToMany(entityName, pk, collections) {
        if (!this.metadata[entityName]) {
            return;
        }
        const props = this.metadata[entityName].properties;
        const owners = Object.keys(collections).filter(k => props[k].owner);
        for (const k of owners) {
            const prop = props[k];
            const fk1 = prop.joinColumn;
            const fk2 = prop.inverseJoinColumn;
            const qb1 = this.createQueryBuilder(prop.pivotTable);
            await qb1.delete({ [fk1]: pk }).execute('run', false);
            for (const item of collections[k]) {
                const qb2 = this.createQueryBuilder(prop.pivotTable);
                await qb2.insert({ [fk1]: pk, [fk2]: item }).execute('run', false);
            }
        }
    }
}
exports.AbstractSqlDriver = AbstractSqlDriver;
