"use strict";
Object.defineProperty(exports, "__esModule", { value: true });
const metadata_1 = require("../metadata");
const enums_1 = require("./enums");
class ArrayCollection {
    constructor(owner, items) {
        this.owner = owner;
        this.items = [];
        if (items) {
            this.items = items;
            Object.assign(this, items);
        }
    }
    getItems() {
        return [...this.items];
    }
    toArray() {
        return this.getItems().map(item => {
            const meta = metadata_1.MetadataStorage.getMetadata(item.constructor.name);
            const args = [...meta.toJsonParams.map(() => undefined), [this.property.name]];
            return item.toJSON(...args);
        });
    }
    getIdentifiers(field) {
        const items = this.getItems();
        if (items.length === 0) {
            return [];
        }
        field = field || this.items[0].__serializedPrimaryKeyField;
        return this.getItems().map(i => i[field]);
    }
    add(...items) {
        for (const item of items) {
            if (!this.contains(item)) {
                this.items.push(item);
                this.propagate(item, 'add');
            }
        }
        Object.assign(this, this.items);
    }
    set(items) {
        this.removeAll();
        this.add(...items);
    }
    remove(...items) {
        for (const item of items) {
            const idx = this.items.findIndex(i => i.__serializedPrimaryKey === item.__serializedPrimaryKey);
            if (idx !== -1) {
                delete this[this.items.length - 1]; // remove last item
                this.items.splice(idx, 1);
                Object.assign(this, this.items); // reassign array access
            }
            this.propagate(item, 'remove');
        }
    }
    removeAll() {
        this.remove(...this.items);
    }
    contains(item) {
        return !!this.items.find(i => {
            const objectIdentity = i === item;
            const primaryKeyIdentity = !!i.__primaryKey && !!item.__primaryKey && i.__serializedPrimaryKey === item.__serializedPrimaryKey;
            return objectIdentity || primaryKeyIdentity;
        });
    }
    count() {
        return this.items.length;
    }
    get length() {
        return this.count();
    }
    *[Symbol.iterator]() {
        for (const item of this.items) {
            yield item;
        }
    }
    propagate(item, method) {
        if (this.property.owner && this.property.inversedBy) {
            this.propagateToInverseSide(item, method);
        }
        else if (!this.property.owner && this.property.mappedBy) {
            this.propagateToOwningSide(item, method);
        }
    }
    propagateToInverseSide(item, method) {
        const collection = item[this.property.inversedBy];
        if (this.shouldPropagateToCollection(collection, method)) {
            collection[method](this.owner);
        }
    }
    propagateToOwningSide(item, method) {
        const collection = item[this.property.mappedBy];
        if (this.property.reference === enums_1.ReferenceType.MANY_TO_MANY && this.shouldPropagateToCollection(collection, method)) {
            collection[method](this.owner);
        }
        else if (this.property.reference === enums_1.ReferenceType.ONE_TO_MANY) {
            const value = method === 'add' ? this.owner : null;
            item[this.property.mappedBy] = value;
        }
    }
    shouldPropagateToCollection(collection, method) {
        if (!collection.isInitialized()) {
            return false;
        }
        if (method === 'add') {
            return !collection.contains(this.owner);
        }
        // remove
        return collection.contains(this.owner);
    }
    get property() {
        if (!this._property) {
            const meta = metadata_1.MetadataStorage.getMetadata(this.owner.constructor.name);
            const field = Object.keys(meta.properties).find(k => this.owner[k] === this);
            this._property = meta.properties[field];
        }
        return this._property;
    }
}
exports.ArrayCollection = ArrayCollection;
