"use strict";
Object.defineProperty(exports, "__esModule", { value: true });
const __1 = require("..");
const ArrayCollection_1 = require("./ArrayCollection");
const enums_1 = require("./enums");
class Collection extends ArrayCollection_1.ArrayCollection {
    constructor(owner, items, initialized = true) {
        super(owner, items);
        this.initialized = false;
        this.dirty = false;
        this._populated = false;
        this.initialized = !!items || initialized;
    }
    getItems() {
        this.checkInitialized();
        return super.getItems();
    }
    add(...items) {
        items = items.map(item => this.mapToEntity(item));
        this.modify('add', items);
        for (const item of items) {
            this.owner.__em.getUnitOfWork().cancelOrphanRemoval(item);
        }
    }
    set(items, initialize = false) {
        if (initialize) {
            this.initialized = true;
        }
        items = items.map(item => this.mapToEntity(item));
        super.set(items);
        this.dirty = !initialize;
        for (const item of items) {
            this.owner.__em.getUnitOfWork().cancelOrphanRemoval(item);
        }
    }
    remove(...items) {
        this.modify('remove', items);
        if (this.property.orphanRemoval) {
            for (const item of items) {
                this.owner.__em.getUnitOfWork().scheduleOrphanRemoval(item);
            }
        }
    }
    contains(item) {
        this.checkInitialized();
        return super.contains(item);
    }
    count() {
        this.checkInitialized();
        return super.count();
    }
    isInitialized(fully = false) {
        if (fully) {
            return this.initialized && this.items.every(item => item.isInitialized());
        }
        return this.initialized;
    }
    shouldPopulate() {
        return this._populated;
    }
    populated(populated = true) {
        this._populated = populated;
    }
    isDirty() {
        return this.dirty;
    }
    setDirty(dirty = true) {
        this.dirty = dirty && this.property.owner; // set dirty flag only to owning side
    }
    async init(populate = []) {
        const em = this.owner.__em;
        if (!this.initialized && this.property.reference === enums_1.ReferenceType.MANY_TO_MANY && em.getDriver().getPlatform().usesPivotTable()) {
            const map = await em.getDriver().loadFromPivotTable(this.property, [this.owner.__primaryKey]);
            this.set(map[this.owner.__primaryKey].map(item => em.merge(this.property.type, item)), true);
            return this;
        }
        // do not make db call if we know we will get no results
        if (this.property.reference === enums_1.ReferenceType.MANY_TO_MANY && (this.property.owner || em.getDriver().getPlatform().usesPivotTable()) && this.length === 0) {
            this.initialized = true;
            this.dirty = false;
            this.populated();
            return this;
        }
        const { cond, orderBy } = this.createCondition();
        const order = [...this.items]; // copy order of references
        this.items.length = 0;
        const items = await em.find(this.property.type, cond, populate, orderBy);
        this.reorderItems(items, order);
        this.items.push(...items);
        Object.assign(this, items);
        this.initialized = true;
        this.dirty = false;
        this.populated();
        return this;
    }
    createCondition() {
        const cond = {};
        let orderBy;
        if (this.property.reference === enums_1.ReferenceType.ONE_TO_MANY) {
            cond[this.property.mappedBy] = this.owner.__primaryKey;
            orderBy = this.property.orderBy || { [this.property.referenceColumnName]: __1.QueryOrder.ASC };
        }
        else { // MANY_TO_MANY
            this.createManyToManyCondition(cond);
        }
        return { cond, orderBy };
    }
    createManyToManyCondition(cond) {
        if (this.property.owner || this.owner.__em.getDriver().getPlatform().usesPivotTable()) {
            const pk = this.items[0].__primaryKeyField; // we know there is at least one item as it was checked in init method
            cond[pk] = { $in: this.items.map(item => item.__primaryKey) };
        }
        else {
            cond[this.property.mappedBy] = this.owner.__primaryKey;
        }
    }
    modify(method, items) {
        this.checkInitialized();
        super[method](...items);
        this.setDirty();
    }
    checkInitialized() {
        if (!this.isInitialized()) {
            throw new Error(`Collection ${this.property.type}[] of entity ${this.owner.constructor.name}[${this.owner.__primaryKey}] not initialized`);
        }
    }
    /**
     * re-orders items after searching with `$in` operator
     */
    reorderItems(items, order) {
        if (this.property.reference === enums_1.ReferenceType.MANY_TO_MANY && this.property.owner) {
            items.sort((a, b) => order.indexOf(a) - order.indexOf(b));
        }
    }
    mapToEntity(item) {
        if (__1.Utils.isEntity(item)) {
            return item;
        }
        const em = this.owner.__em;
        let entity;
        if (__1.Utils.isPrimaryKey(item)) {
            entity = em.getReference(this.property.type, item);
        }
        else {
            entity = em.create(this.property.type, item);
        }
        if (entity.__primaryKey) {
            return em.merge(this.property.type, entity);
        }
        return entity;
    }
}
exports.Collection = Collection;
