"use strict";
Object.defineProperty(exports, "__esModule", { value: true });
const EntityFactory_1 = require("./EntityFactory");
const utils_1 = require("../utils");
const metadata_1 = require("../metadata");
const enums_1 = require("./enums");
class EntityAssigner {
    static assign(entity, data, onlyProperties = false) {
        const meta = metadata_1.MetadataStorage.getMetadata(entity.constructor.name);
        const props = meta.properties;
        const options = (typeof onlyProperties === 'boolean' ? { onlyProperties } : onlyProperties);
        Object.keys(data).forEach(prop => {
            if (options.onlyProperties && !(prop in props)) {
                return;
            }
            const value = data[prop];
            if (props[prop] && [enums_1.ReferenceType.MANY_TO_ONE, enums_1.ReferenceType.ONE_TO_ONE].includes(props[prop].reference) && value) {
                return EntityAssigner.assignReference(entity, value, props[prop], entity.__em);
            }
            if (props[prop] && utils_1.Utils.isCollection(entity[prop], props[prop]) && Array.isArray(value)) {
                return EntityAssigner.assignCollection(entity, entity[prop], value, props[prop], entity.__em);
            }
            if (props[prop] && props[prop].reference === enums_1.ReferenceType.SCALAR && EntityFactory_1.SCALAR_TYPES.includes(props[prop].type)) {
                return entity[prop] = entity.__em.getValidator().validateProperty(props[prop], value, entity);
            }
            if (options.mergeObjects && utils_1.Utils.isObject(value)) {
                utils_1.Utils.merge(entity[prop], value);
            }
            else {
                entity[prop] = value;
            }
        });
    }
    static assignReference(entity, value, prop, em) {
        if (utils_1.Utils.isEntity(value)) {
            entity[prop.name] = value;
            return;
        }
        if (utils_1.Utils.isPrimaryKey(value)) {
            entity[prop.name] = em.getReference(prop.type, value);
            return;
        }
        if (utils_1.Utils.isObject(value)) {
            entity[prop.name] = em.create(prop.type, value);
            return;
        }
        const name = entity.constructor.name;
        throw new Error(`Invalid reference value provided for '${name}.${prop.name}' in ${name}.assign(): ${JSON.stringify(value)}`);
    }
    static assignCollection(entity, collection, value, prop, em) {
        const invalid = [];
        const items = value.map((item) => this.createCollectionItem(item, em, prop, invalid));
        if (invalid.length > 0) {
            const name = entity.constructor.name;
            throw new Error(`Invalid collection values provided for '${name}.${prop.name}' in ${name}.assign(): ${JSON.stringify(invalid)}`);
        }
        collection.set(items, true);
        collection.setDirty();
    }
    static createCollectionItem(item, em, prop, invalid) {
        if (utils_1.Utils.isEntity(item)) {
            return item;
        }
        if (utils_1.Utils.isPrimaryKey(item)) {
            return em.getReference(prop.type, item);
        }
        if (utils_1.Utils.isObject(item)) {
            return em.create(prop.type, item);
        }
        invalid.push(item);
        return item;
    }
}
exports.EntityAssigner = EntityAssigner;
