"use strict";
var __importDefault = (this && this.__importDefault) || function (mod) {
    return (mod && mod.__esModule) ? mod : { "default": mod };
};
Object.defineProperty(exports, "__esModule", { value: true });
const globby_1 = __importDefault(require("globby"));
const path_1 = require("path");
const utils_1 = require("../utils");
const MetadataValidator_1 = require("./MetadataValidator");
const MetadataStorage_1 = require("./MetadataStorage");
const entity_1 = require("../entity");
class MetadataDiscovery {
    constructor(em, config, logger) {
        this.em = em;
        this.config = config;
        this.logger = logger;
        this.metadata = MetadataStorage_1.MetadataStorage.getMetadata();
        this.namingStrategy = this.config.getNamingStrategy();
        this.metadataProvider = this.config.getMetadataProvider();
        this.cache = this.config.getCacheAdapter();
        this.platform = this.em.getDriver().getPlatform();
        this.validator = new MetadataValidator_1.MetadataValidator();
        this.discovered = [];
    }
    async discover() {
        const startTime = Date.now();
        this.logger.debug(`ORM entity discovery started`);
        this.discovered.length = 0;
        const tsNode = process.argv[0].endsWith('ts-node') || process.argv.slice(1).some(arg => arg.includes('ts-node'));
        if (this.config.get('entities').length > 0) {
            await utils_1.Utils.runSerial(this.config.get('entities'), entity => this.discoverEntity(entity));
        }
        else if (tsNode) {
            await utils_1.Utils.runSerial(this.config.get('entitiesDirsTs'), dir => this.discoverDirectory(dir));
        }
        else {
            await utils_1.Utils.runSerial(this.config.get('entitiesDirs'), dir => this.discoverDirectory(dir));
        }
        // validate base entities
        this.discovered
            .filter(meta => meta.extends && !this.discovered.find(m => m.prototype.constructor.name === meta.extends))
            .forEach(meta => { throw utils_1.ValidationError.fromUnknownBaseEntity(meta); });
        // ignore base entities (not annotated with @Entity)
        const filtered = this.discovered.filter(meta => meta.name);
        filtered.forEach(meta => this.defineBaseEntityProperties(meta));
        filtered.forEach(meta => this.discovered.push(...this.processEntity(meta)));
        const diff = Date.now() - startTime;
        this.logger.debug(`- entity discovery finished after ${diff} ms`);
        const discovered = {};
        this.discovered
            .filter(meta => meta.name)
            .forEach(meta => discovered[meta.name] = meta);
        return discovered;
    }
    async discoverDirectory(basePath) {
        const files = await globby_1.default('*', { cwd: utils_1.Utils.normalizePath(this.config.get('baseDir'), basePath) });
        this.logger.debug(`- processing ${files.length} files from directory ${basePath}`);
        for (const file of files) {
            if (!file.match(/\.[jt]s$/) ||
                file.endsWith('.js.map') ||
                file.endsWith('.d.ts') ||
                file.startsWith('.') ||
                file.match(/index\.[jt]s$/)) {
                this.logger.debug(`- ignoring file ${file}`);
                continue;
            }
            const name = this.getClassName(file);
            const path = utils_1.Utils.normalizePath(this.config.get('baseDir'), basePath, file);
            const target = require(path)[name]; // include the file to trigger loading of metadata
            await this.discoverEntity(target, path);
        }
    }
    async discoverEntity(entity, path) {
        entity = this.metadataProvider.prepare(entity);
        this.logger.debug(`- processing entity ${entity.name}`);
        const meta = MetadataStorage_1.MetadataStorage.getMetadata(entity.name);
        meta.prototype = entity.prototype;
        meta.path = path || meta.path;
        meta.toJsonParams = utils_1.Utils.getParamNames(entity.prototype.toJSON || '').filter(p => p !== '...args');
        const cache = meta.path && await this.cache.get(entity.name + path_1.extname(meta.path));
        if (cache) {
            this.logger.debug(`- using cached metadata for entity ${entity.name}`);
            this.metadataProvider.loadFromCache(meta, cache);
            this.discovered.push(meta);
            return;
        }
        await this.metadataProvider.loadEntityMetadata(meta, entity.name);
        if (!meta.collection && meta.name) {
            meta.collection = this.namingStrategy.classToTableName(meta.name);
        }
        await this.saveToCache(meta, entity);
        this.discovered.push(meta);
    }
    async saveToCache(meta, entity) {
        const copy = Object.assign({}, meta);
        delete copy.prototype;
        // base entity without properties might not have path, but nothing to cache there
        if (meta.path) {
            await this.cache.set(entity.name + path_1.extname(meta.path), copy, meta.path);
        }
    }
    applyNamingStrategy(meta, prop) {
        if (!prop.fieldName) {
            this.initFieldName(prop);
        }
        if (prop.reference === entity_1.ReferenceType.MANY_TO_MANY) {
            this.initManyToManyFields(meta, prop);
        }
        if (prop.reference === entity_1.ReferenceType.MANY_TO_ONE) {
            const meta2 = this.metadata[prop.type];
            prop.inverseJoinColumn = meta2.properties[meta2.primaryKey].fieldName;
        }
        if (prop.reference === entity_1.ReferenceType.ONE_TO_MANY || prop.reference === entity_1.ReferenceType.ONE_TO_ONE) {
            this.initOneToManyFields(meta, prop);
        }
    }
    initFieldName(prop) {
        if (prop.reference === entity_1.ReferenceType.SCALAR) {
            prop.fieldName = this.namingStrategy.propertyToColumnName(prop.name);
        }
        else if (prop.reference === entity_1.ReferenceType.MANY_TO_ONE || prop.reference === entity_1.ReferenceType.ONE_TO_ONE) {
            prop.fieldName = this.initManyToOneFieldName(prop, prop.name);
        }
        else if (prop.reference === entity_1.ReferenceType.MANY_TO_MANY && prop.owner) {
            prop.fieldName = this.namingStrategy.propertyToColumnName(prop.name);
        }
    }
    initManyToOneFieldName(prop, name) {
        const meta2 = this.metadata[prop.type];
        const referenceColumnName = meta2.properties[meta2.primaryKey].fieldName;
        return this.namingStrategy.joinKeyColumnName(name, referenceColumnName);
    }
    initManyToManyFields(meta, prop) {
        if (!prop.pivotTable && prop.owner) {
            prop.pivotTable = this.namingStrategy.joinTableName(meta.name, prop.type, prop.name);
        }
        if (!prop.referenceColumnName) {
            prop.referenceColumnName = meta.properties[meta.primaryKey].fieldName;
        }
        if (!prop.inverseJoinColumn) {
            prop.inverseJoinColumn = this.initManyToOneFieldName(prop, prop.type);
        }
        if (!prop.joinColumn) {
            prop.joinColumn = this.namingStrategy.joinKeyColumnName(meta.name, prop.referenceColumnName);
        }
    }
    initOneToManyFields(meta, prop) {
        if (!prop.joinColumn) {
            prop.joinColumn = this.namingStrategy.joinColumnName(prop.name);
        }
        if (prop.reference === entity_1.ReferenceType.ONE_TO_ONE && !prop.inverseJoinColumn && prop.mappedBy) {
            prop.inverseJoinColumn = this.metadata[prop.type].properties[prop.mappedBy].fieldName;
        }
        if (!prop.referenceColumnName) {
            prop.referenceColumnName = meta.properties[meta.primaryKey].fieldName;
        }
    }
    processEntity(meta) {
        this.defineBaseEntityProperties(meta);
        // BC with 1:m `fk` option
        Object.values(meta.properties)
            .filter(prop => prop.reference === entity_1.ReferenceType.ONE_TO_MANY)
            .forEach(prop => utils_1.Utils.renameKey(prop, 'fk', 'mappedBy'));
        this.validator.validateEntityDefinition(this.metadata, meta.name);
        Object.values(meta.properties).forEach(prop => {
            this.applyNamingStrategy(meta, prop);
            if (prop.version) {
                meta.versionProperty = prop.name;
                prop.default = this.getDefaultVersionValue(prop);
            }
        });
        meta.serializedPrimaryKey = this.platform.getSerializedPrimaryKeyField(meta.primaryKey);
        if (!utils_1.Utils.isEntity(meta.prototype)) {
            entity_1.EntityHelper.decorate(meta, this.em);
        }
        const ret = [];
        if (this.platform.usesPivotTable()) {
            Object.values(meta.properties).forEach(prop => {
                const pivotMeta = this.definePivotTableEntity(meta, prop);
                if (pivotMeta) {
                    ret.push(pivotMeta);
                }
            });
        }
        return ret;
    }
    definePivotTableEntity(meta, prop) {
        if (prop.reference === entity_1.ReferenceType.MANY_TO_MANY && prop.owner && prop.pivotTable) {
            const pk = this.namingStrategy.referenceColumnName();
            const primaryProp = { name: pk, type: 'number', reference: entity_1.ReferenceType.SCALAR, primary: true };
            this.initFieldName(primaryProp);
            return this.metadata[prop.pivotTable] = {
                name: prop.pivotTable,
                collection: prop.pivotTable,
                primaryKey: pk,
                properties: {
                    [pk]: primaryProp,
                    [meta.name]: this.definePivotProperty(prop, meta.name, prop.type),
                    [prop.type]: this.definePivotProperty(prop, prop.type, meta.name),
                },
            };
        }
    }
    definePivotProperty(prop, name, inverse) {
        const ret = { name, type: name, reference: entity_1.ReferenceType.MANY_TO_ONE, cascade: [entity_1.Cascade.ALL] };
        if (name === prop.type) {
            const meta = this.metadata[name];
            const prop2 = meta.properties[meta.primaryKey];
            if (!prop2.fieldName) {
                this.initFieldName(prop2);
            }
            ret.owner = false;
            ret.mappedBy = inverse;
            ret.referenceColumnName = prop2.fieldName;
            ret.fieldName = ret.joinColumn = prop.inverseJoinColumn;
            ret.inverseJoinColumn = prop2.fieldName;
        }
        else {
            ret.owner = true;
            ret.inversedBy = inverse;
            ret.referenceColumnName = prop.referenceColumnName;
            ret.fieldName = ret.joinColumn = prop.joinColumn;
            ret.inverseJoinColumn = prop.referenceColumnName;
        }
        return ret;
    }
    getClassName(file) {
        const name = file.split('.')[0];
        const ret = name.replace(/-(\w)/, m => m[1].toUpperCase());
        return ret.charAt(0).toUpperCase() + ret.slice(1);
    }
    defineBaseEntityProperties(meta) {
        const base = this.metadata[meta.extends];
        if (!meta.extends || !base) {
            return;
        }
        meta.properties = Object.assign({}, base.properties, meta.properties);
        const primary = Object.values(meta.properties).find(p => p.primary);
        if (primary && !meta.primaryKey) {
            meta.primaryKey = primary.name;
        }
    }
    getDefaultVersionValue(prop) {
        if (prop.default) {
            return prop.default;
        }
        if (prop.type.toLowerCase() === 'date') {
            prop.length = prop.length || 3;
            return this.platform.getCurrentTimestampSQL(prop.length);
        }
        return '1';
    }
}
exports.MetadataDiscovery = MetadataDiscovery;
