"use strict";
Object.defineProperty(exports, "__esModule", { value: true });
const utils_1 = require("../utils");
const QueryBuilderHelper_1 = require("./QueryBuilderHelper");
const SmartQueryHelper_1 = require("./SmartQueryHelper");
const entity_1 = require("../entity");
const enums_1 = require("./enums");
const unit_of_work_1 = require("../unit-of-work");
/**
 * SQL query builder
 */
class QueryBuilder {
    constructor(entityName, metadata, driver, alias = `e0`) {
        this.entityName = entityName;
        this.metadata = metadata;
        this.driver = driver;
        this.alias = alias;
        this.aliasCounter = 1;
        this.flags = [];
        this.finalized = false;
        this._populate = [];
        this._populateMap = {};
        this._joins = {};
        this._aliasMap = {};
        this._cond = {};
        this._orderBy = {};
        this._groupBy = [];
        this._having = {};
        this.connection = this.driver.getConnection();
        this.platform = this.driver.getPlatform();
        this.helper = new QueryBuilderHelper_1.QueryBuilderHelper(this.entityName, this.alias, this._aliasMap, this.metadata, this.platform);
    }
    select(fields, distinct = false) {
        this._fields = utils_1.Utils.asArray(fields);
        if (distinct) {
            this.flags.push(enums_1.QueryFlag.DISTINCT);
        }
        return this.init(enums_1.QueryType.SELECT);
    }
    insert(data) {
        return this.init(enums_1.QueryType.INSERT, data);
    }
    update(data) {
        return this.init(enums_1.QueryType.UPDATE, data);
    }
    delete(cond = {}) {
        return this.init(enums_1.QueryType.DELETE, undefined, cond);
    }
    truncate() {
        return this.init(enums_1.QueryType.TRUNCATE);
    }
    count(field, distinct = false) {
        this.select(field || this.metadata[this.entityName].primaryKey);
        this.flags.push(enums_1.QueryFlag.COUNT);
        if (distinct) {
            this.flags.push(enums_1.QueryFlag.DISTINCT);
        }
        return this;
    }
    join(field, alias, type = 'inner') {
        const [fromAlias, fromField] = this.helper.splitField(field);
        const entityName = this._aliasMap[fromAlias];
        const prop = this.metadata[entityName].properties[fromField];
        this._aliasMap[alias] = prop.type;
        if (prop.reference === entity_1.ReferenceType.ONE_TO_MANY) {
            this._joins[prop.name] = this.helper.joinOneToReference(prop, fromAlias, alias, type);
        }
        else if (prop.reference === entity_1.ReferenceType.MANY_TO_MANY) {
            const pivotAlias = `e${this.aliasCounter++}`;
            const joins = this.helper.joinManyToManyReference(prop, fromAlias, alias, pivotAlias, type);
            this._fields.push(`${pivotAlias}.${prop.name}`);
            Object.assign(this._joins, joins);
        }
        else if (prop.reference === entity_1.ReferenceType.ONE_TO_ONE) {
            this._joins[prop.name] = this.helper.joinOneToReference(prop, fromAlias, alias, type);
        }
        else { // MANY_TO_ONE
            this._joins[prop.name] = this.helper.joinManyToOneReference(prop, fromAlias, alias, type);
        }
        return this;
    }
    leftJoin(field, alias) {
        return this.join(field, alias, 'left');
    }
    where(cond, params, operator) {
        cond = SmartQueryHelper_1.SmartQueryHelper.processWhere(cond, this.entityName);
        if (utils_1.Utils.isString(cond)) {
            cond = { [`(${cond})`]: utils_1.Utils.asArray(params) };
            operator = operator || '$and';
        }
        const op = operator || params;
        const topLevel = !op || Object.keys(this._cond).length === 0;
        if (topLevel) {
            this._cond = this.processWhere(cond);
        }
        else if (Array.isArray(this._cond[op])) {
            this._cond[op].push(this.processWhere(cond));
        }
        else {
            const cond1 = [this._cond, this.processWhere(cond)];
            this._cond = { [op]: cond1 };
        }
        return this;
    }
    andWhere(cond, params) {
        return this.where(cond, params, '$and');
    }
    orWhere(cond, params) {
        return this.where(cond, params, '$or');
    }
    orderBy(orderBy) {
        orderBy = Object.assign({}, orderBy); // copy first
        Object.keys(orderBy).forEach(field => {
            if (!this.metadata[this.entityName] || !this.metadata[this.entityName].properties[field]) {
                return;
            }
            const prop = this.metadata[this.entityName].properties[field];
            utils_1.Utils.renameKey(orderBy, field, prop.fieldName);
        });
        this._orderBy = orderBy;
        return this;
    }
    groupBy(fields) {
        this._groupBy = utils_1.Utils.asArray(fields);
        return this;
    }
    having(cond, params) {
        if (utils_1.Utils.isString(cond)) {
            cond = { [`(${cond})`]: utils_1.Utils.asArray(params) };
        }
        this._having = this.processWhere(cond);
        return this;
    }
    populate(populate) {
        this._populate = populate;
        return this;
    }
    limit(limit, offset = 0) {
        this._limit = limit;
        if (offset) {
            this.offset(offset);
        }
        return this;
    }
    offset(offset) {
        this._offset = offset;
        return this;
    }
    setLockMode(mode) {
        if ([unit_of_work_1.LockMode.NONE, unit_of_work_1.LockMode.PESSIMISTIC_READ, unit_of_work_1.LockMode.PESSIMISTIC_WRITE].includes(mode) && !this.driver.isInTransaction()) {
            throw utils_1.ValidationError.transactionRequired();
        }
        this.lockMode = mode;
        return this;
    }
    getQuery() {
        this.finalize();
        let sql = this.getQueryBase();
        sql += this.helper.getClause('WHERE', this.helper.getQueryCondition(this.type, this._cond).join(' AND '), this._cond);
        sql += this.helper.getClause('GROUP BY', this.prepareFields(this._groupBy), this._groupBy);
        sql += this.helper.getClause('HAVING', this.helper.getQueryCondition(this.type, this._having).join(' AND '), this._having);
        sql += this.helper.getClause('ORDER BY', this.helper.getQueryOrder(this.type, this._orderBy, this._populateMap).join(', '), this._orderBy);
        sql += this.helper.getClause('LIMIT', '?', this._limit);
        sql += this.helper.getClause('OFFSET', '?', this._offset);
        if (this.type === enums_1.QueryType.TRUNCATE && this.platform.usesCascadeStatement()) {
            sql += ' CASCADE';
        }
        sql += this.helper.getLockSQL(this.lockMode);
        return this.helper.finalize(this.type, sql, this.metadata[this.entityName]);
    }
    getParams() {
        this.finalize();
        let ret = [];
        if (this.type === enums_1.QueryType.INSERT && this._data) {
            ret = Object.values(this._data);
        }
        else if (this.type === enums_1.QueryType.UPDATE) {
            ret = Object.values(this._data);
        }
        ret = ret.concat(this.helper.getWhereParams(this._cond));
        ret = ret.concat(this.helper.getWhereParams(this._having));
        if (this._limit) {
            ret.push(this._limit);
        }
        if (this._offset) {
            ret.push(this._offset);
        }
        return SmartQueryHelper_1.SmartQueryHelper.processParams(ret);
    }
    async execute(method = 'all', mapResults = true) {
        const res = await this.connection.execute(this.getQuery(), this.getParams(), method);
        if (!mapResults) {
            return res;
        }
        if (method === 'all' && Array.isArray(res)) {
            return res.map((r) => this.driver.mapResult(r, this.metadata[this.entityName]));
        }
        return this.driver.mapResult(res, this.metadata[this.entityName]);
    }
    clone() {
        const qb = new QueryBuilder(this.entityName, this.metadata, this.driver, this.alias);
        Object.assign(qb, this);
        // clone array/object properties
        const properties = ['flags', '_fields', '_populate', '_populateMap', '_joins', '_aliasMap', '_cond', '_data', '_orderBy'];
        properties.forEach(prop => qb[prop] = utils_1.Utils.copy(this[prop]));
        return qb;
    }
    prepareFields(fields, glue = ', ') {
        const ret = [];
        fields.forEach(f => {
            if (this._joins[f]) {
                ret.push(...this.helper.mapJoinColumns(this.type, this._joins[f]));
                return;
            }
            ret.push(this.helper.mapper(this.type, f));
        });
        Object.keys(this._populateMap).forEach(f => {
            if (!fields.includes(f)) {
                ret.push(...this.helper.mapJoinColumns(this.type, this._joins[f]));
            }
            if (this._joins[f].prop.reference !== entity_1.ReferenceType.ONE_TO_ONE) {
                utils_1.Utils.renameKey(this._cond, this._joins[f].inverseJoinColumn, `${this._joins[f].alias}.${this._joins[f].inverseJoinColumn}`);
            }
        });
        if (this.flags.includes(enums_1.QueryFlag.COUNT)) {
            if (this.flags.includes(enums_1.QueryFlag.DISTINCT)) {
                return `COUNT(DISTINCT ${ret[0]}) AS ${this.helper.wrap('count')}`;
            }
            return `COUNT(${ret[0]}) AS ${this.helper.wrap('count')}`;
        }
        return ret.join(glue);
    }
    processWhere(cond) {
        cond = Object.assign({}, cond); // copy first
        Object.keys(cond).forEach(field => {
            this.helper.replaceEmptyInConditions(cond, field);
            if (!this.metadata[this.entityName] || !this.metadata[this.entityName].properties[field]) {
                return;
            }
            const prop = this.metadata[this.entityName].properties[field];
            if (prop.reference === entity_1.ReferenceType.MANY_TO_MANY) {
                this.processManyToMany(prop, cond);
            }
            else if (prop.reference === entity_1.ReferenceType.ONE_TO_MANY) {
                this.processOneToMany(prop, cond);
            }
            else if (prop.reference === entity_1.ReferenceType.ONE_TO_ONE) {
                this.processOneToOne(prop, cond);
            }
            else {
                utils_1.Utils.renameKey(cond, field, prop.fieldName);
            }
        });
        return cond;
    }
    processOneToOne(prop, cond) {
        if (prop.owner) {
            return utils_1.Utils.renameKey(cond, prop.name, prop.fieldName);
        }
        this._fields.push(prop.name);
        const alias2 = `e${this.aliasCounter++}`;
        this._joins[prop.name] = this.helper.joinOneToReference(prop, this.alias, alias2, 'left');
        const prop2 = this.metadata[prop.type].properties[prop.mappedBy];
        utils_1.Utils.renameKey(cond, prop.name, `${alias2}.${prop2.referenceColumnName}`);
    }
    processManyToMany(prop, cond) {
        const alias1 = `e${this.aliasCounter++}`;
        const join = {
            type: 'left',
            alias: alias1,
            ownerAlias: this.alias,
            joinColumn: prop.joinColumn,
            inverseJoinColumn: prop.inverseJoinColumn,
            primaryKey: prop.referenceColumnName,
        };
        if (prop.owner) {
            this._joins[prop.name] = Object.assign(join, { table: prop.pivotTable });
        }
        else {
            const prop2 = this.metadata[prop.type].properties[prop.mappedBy];
            this._joins[prop.name] = Object.assign(join, { table: prop2.pivotTable });
        }
        this._fields.push(prop.name);
        utils_1.Utils.renameKey(cond, prop.name, `${alias1}.${prop.inverseJoinColumn}`);
    }
    processOneToMany(prop, cond) {
        const alias2 = `e${this.aliasCounter++}`;
        this._joins[prop.name] = this.helper.joinOneToReference(prop, this.alias, alias2, 'left');
        utils_1.Utils.renameKey(cond, prop.name, `${alias2}.${prop.referenceColumnName}`);
    }
    init(type, data, cond) {
        this.type = type;
        this._aliasMap[this.alias] = this.entityName;
        if (data) {
            this._data = this.helper.processData(data);
        }
        if (cond) {
            this._cond = this.processWhere(cond);
        }
        return this;
    }
    getQueryBase() {
        let sql = this.type + ' ';
        switch (this.type) {
            case enums_1.QueryType.SELECT:
                sql += this.flags.includes(enums_1.QueryFlag.DISTINCT) && !this.flags.includes(enums_1.QueryFlag.COUNT) ? 'DISTINCT ' : '';
                sql += this.prepareFields(this._fields);
                sql += ` FROM ${this.helper.getTableName(this.entityName, true)} AS ${this.helper.wrap(this.alias)}`;
                sql += this.helper.processJoins(this._joins);
                break;
            case enums_1.QueryType.INSERT:
                sql += `INTO ${this.helper.getTableName(this.entityName, true)}`;
                sql += ' (' + Object.keys(this._data).map(k => this.helper.wrap(k)).join(', ') + ')';
                sql += ' VALUES (' + Object.keys(this._data).map(() => '?').join(', ') + ')';
                break;
            case enums_1.QueryType.UPDATE:
                sql += this.helper.getTableName(this.entityName, true);
                const set = Object.keys(this._data).map(k => this.helper.wrap(k) + ' = ?');
                this.helper.updateVersionProperty(set);
                sql += ' SET ' + set.join(', ');
                break;
            case enums_1.QueryType.DELETE:
                sql += 'FROM ' + this.helper.getTableName(this.entityName, true);
                break;
            case enums_1.QueryType.TRUNCATE:
                sql += 'TABLE ' + this.helper.getTableName(this.entityName, true);
                break;
        }
        return sql;
    }
    finalize() {
        if (this.finalized) {
            return;
        }
        this._populate.forEach(field => {
            if (this._joins[field]) {
                return this._populateMap[field] = this._joins[field].alias;
            }
            if (this.metadata[field]) { // pivot table entity
                const prop = this.metadata[field].properties[this.entityName];
                this._joins[field] = this.helper.joinPivotTable(field, prop, this.alias, `e${this.aliasCounter++}`, 'left');
                this._populateMap[field] = this._joins[field].alias;
            }
            else if (this.helper.isOneToOneInverse(field)) {
                const prop = this.metadata[this.entityName].properties[field];
                this._joins[prop.name] = this.helper.joinOneToReference(prop, this.alias, `e${this.aliasCounter++}`, 'left');
                this._populateMap[field] = this._joins[field].alias;
            }
        });
        this.finalized = true;
    }
}
exports.QueryBuilder = QueryBuilder;
