"use strict";
Object.defineProperty(exports, "__esModule", { value: true });
const utils_1 = require("../utils");
const enums_1 = require("./enums");
const entity_1 = require("../entity");
const unit_of_work_1 = require("../unit-of-work");
class QueryBuilderHelper {
    constructor(entityName, alias, aliasMap, metadata, platform) {
        this.entityName = entityName;
        this.alias = alias;
        this.aliasMap = aliasMap;
        this.metadata = metadata;
        this.platform = platform;
        this.quoteChar = this.platform.getSchemaHelper().getIdentifierQuoteCharacter();
    }
    getWhereParams(conditions) {
        const ret = [];
        Object.entries(conditions).forEach(([key, cond]) => {
            if (['$and', '$or', '$not'].includes(key)) {
                return ret.push(...this.getGroupWhereParams(key, cond));
            }
            if (cond === null) {
                return;
            }
            // grouped condition for one field
            if (utils_1.Utils.isObject(cond) && Object.keys(cond).length > 1) {
                const subConditions = Object.entries(cond).map(([subKey, subValue]) => ({ [key]: { [subKey]: subValue } }));
                return ret.push(...this.getWhereParams({ $and: subConditions }));
            }
            if (cond instanceof RegExp) {
                return ret.push(this.getRegExpParam(cond));
            }
            if (!utils_1.Utils.isObject(cond) && !Array.isArray(cond)) {
                return ret.push(cond);
            }
            ret.push(...this.processComplexParam(key, cond));
        });
        return ret;
    }
    wrap(field) {
        if (field === '*') {
            return field;
        }
        return this.quoteChar + field + this.quoteChar;
    }
    mapper(type, field, value, alias) {
        let ret = field;
        const customExpression = field.match(/\(.*\)| |^\d/);
        // do not wrap custom expressions
        if (!customExpression) {
            ret = this.prefixAndWrap(field);
        }
        if (typeof value !== 'undefined') {
            ret += this.processValue(field, value);
        }
        if (alias) {
            ret += ' AS ' + this.wrap(alias);
        }
        if (type !== enums_1.QueryType.SELECT || customExpression || this.isPrefixed(ret)) {
            return ret;
        }
        return this.wrap(this.alias) + '.' + ret;
    }
    processData(data) {
        data = Object.assign({}, data); // copy first
        Object.keys(data).forEach(k => {
            if (this.metadata[this.entityName] && this.metadata[this.entityName].properties[k]) {
                const prop = this.metadata[this.entityName].properties[k];
                if (Array.isArray(data[k])) {
                    data[k] = JSON.stringify(data[k]);
                }
                if (prop.fieldName) {
                    utils_1.Utils.renameKey(data, k, prop.fieldName);
                }
            }
        });
        return data;
    }
    joinOneToReference(prop, ownerAlias, alias, type) {
        const prop2 = this.metadata[prop.type].properties[prop.mappedBy || prop.inversedBy];
        return {
            table: this.getTableName(prop.type),
            joinColumn: prop.owner ? prop2.referenceColumnName : prop2.fieldName,
            inverseJoinColumn: prop2.referenceColumnName,
            primaryKey: prop.owner ? prop.joinColumn : prop.referenceColumnName,
            ownerAlias,
            alias,
            prop,
            type,
        };
    }
    joinManyToOneReference(prop, ownerAlias, alias, type) {
        return {
            table: this.getTableName(prop.type),
            joinColumn: prop.inverseJoinColumn,
            primaryKey: prop.fieldName,
            ownerAlias,
            alias,
            prop,
            type,
        };
    }
    joinManyToManyReference(prop, ownerAlias, alias, pivotAlias, type) {
        const join = {
            type,
            ownerAlias,
            alias: pivotAlias,
            joinColumn: prop.joinColumn,
            inverseJoinColumn: prop.inverseJoinColumn,
            primaryKey: prop.referenceColumnName,
        };
        const name = `${pivotAlias}.${prop.name}`;
        const ret = {};
        if (prop.owner) {
            ret[name] = Object.assign(join, { table: prop.pivotTable });
        }
        else {
            const prop2 = this.metadata[prop.type].properties[prop.mappedBy];
            ret[name] = Object.assign(join, { table: prop2.pivotTable });
        }
        if (prop.owner) {
            const prop2 = this.metadata[prop.pivotTable].properties[prop.type];
            ret[prop2.name] = this.joinManyToOneReference(prop2, pivotAlias, alias, type);
        }
        else {
            const prop2 = this.metadata[prop.type].properties[prop.mappedBy];
            const prop3 = this.metadata[prop2.pivotTable].properties[prop.type];
            ret[prop3.name] = this.joinManyToOneReference(prop3, pivotAlias, alias, type);
        }
        return ret;
    }
    joinPivotTable(field, prop, ownerAlias, alias, type) {
        const prop2 = this.metadata[field].properties[prop.mappedBy || prop.inversedBy];
        return {
            table: this.metadata[field].collection,
            joinColumn: prop.joinColumn,
            inverseJoinColumn: prop2.joinColumn,
            primaryKey: prop.referenceColumnName,
            ownerAlias,
            alias,
            prop,
            type,
        };
    }
    processJoins(joins) {
        return Object.values(joins).map(join => {
            const type = join.type === 'inner' ? '' : join.type.toUpperCase() + ' ';
            return ` ${type}JOIN ${this.wrap(join.table)} AS ${this.wrap(join.alias)} ON ${this.wrap(join.ownerAlias)}.${this.wrap(join.primaryKey)} = ${this.wrap(join.alias)}.${this.wrap(join.joinColumn)}`;
        }).join('');
    }
    mapJoinColumns(type, join) {
        if (join.prop && join.prop.reference === entity_1.ReferenceType.ONE_TO_ONE && !join.prop.owner) {
            return [this.mapper(type, `${join.alias}.${join.inverseJoinColumn}`, undefined, join.prop.fieldName)];
        }
        return [
            this.mapper(type, `${join.alias}.${join.joinColumn}`),
            this.mapper(type, `${join.alias}.${join.inverseJoinColumn}`),
        ];
    }
    isOneToOneInverse(field) {
        const prop = this.metadata[this.entityName] && this.metadata[this.entityName].properties[field];
        return prop && prop.reference === entity_1.ReferenceType.ONE_TO_ONE && !prop.owner;
    }
    getTableName(entityName, wrap = false) {
        const name = this.metadata[entityName] ? this.metadata[entityName].collection : entityName;
        if (wrap) {
            return this.wrap(name);
        }
        return name;
    }
    getRegExpParam(re) {
        const value = re.source
            .replace(/\.\*/g, '%') // .* -> %
            .replace(/\./g, '_') // .  -> _
            .replace(/\\_/g, '.') // \. -> .
            .replace(/^\^/g, '') // remove ^ from start
            .replace(/\$$/g, ''); // remove $ from end
        if (re.source.startsWith('^') && re.source.endsWith('$')) {
            return value;
        }
        if (re.source.startsWith('^')) {
            return value + '%';
        }
        if (re.source.endsWith('$')) {
            return '%' + value;
        }
        return `%${value}%`;
    }
    getQueryCondition(type, cond) {
        return Object.keys(cond).map(k => {
            if (k === '$and' || k === '$or') {
                return this.getGroupQueryCondition(type, k, cond[k]);
            }
            if (k === '$not') {
                return 'NOT (' + this.getQueryCondition(type, cond[k])[0] + ')';
            }
            // grouped condition for one field
            if (utils_1.Utils.isObject(cond[k]) && Object.keys(cond[k]).length > 1) {
                const subCondition = Object.entries(cond[k]).map(([subKey, subValue]) => ({ [k]: { [subKey]: subValue } }));
                return this.getGroupQueryCondition(type, '$and', subCondition);
            }
            return this.mapper(type, k, cond[k]);
        });
    }
    getQueryOrder(type, orderBy, populate) {
        return Object.keys(orderBy).map(k => {
            let alias = this.alias;
            let field = k;
            if (k.includes('.')) {
                [alias, field] = k.split('.');
            }
            alias = populate[alias] || alias;
            const direction = orderBy[k];
            const order = utils_1.Utils.isNumber(direction) ? enums_1.QueryOrderNumeric[direction] : direction;
            return this.mapper(type, `${alias}.${field}`) + ' ' + order;
        });
    }
    getClause(type, clause, data) {
        if (utils_1.Utils.isEmpty(data)) {
            return '';
        }
        return ` ${type} ${clause}`;
    }
    finalize(type, sql, meta) {
        let append = '';
        const useReturningStatement = type === enums_1.QueryType.INSERT && this.platform.usesReturningStatement();
        if (useReturningStatement && meta) {
            const returningProps = Object.values(meta.properties).filter(prop => prop.primary || prop.default);
            append = ` RETURNING ${returningProps.map(prop => this.wrap(prop.fieldName)).join(', ')}`;
        }
        if (this.platform.getParameterPlaceholder() === '?') {
            return sql + append;
        }
        let index = 1;
        return sql.replace(/(\?)/g, () => {
            return this.platform.getParameterPlaceholder(index++);
        }) + append;
    }
    splitField(field) {
        const [a, b] = field.split('.');
        const fromAlias = b ? a : this.alias;
        const fromField = b || a;
        return [fromAlias, fromField];
    }
    replaceEmptyInConditions(cond, field) {
        if (QueryBuilderHelper.GROUP_OPERATORS[field]) {
            cond[field].forEach((subCond) => Object.keys(subCond).forEach(key => this.replaceEmptyInConditions(subCond, key)));
            cond[field] = cond[field].filter((subCond) => !utils_1.Utils.isObject(subCond) || Object.keys(subCond).length > 0);
            return;
        }
        if (!utils_1.Utils.isObject(cond[field]) || cond[field] instanceof RegExp) {
            return;
        }
        // IN () is always false
        if (cond[field] && cond[field].$in && cond[field].$in.length === 0) {
            cond[field].$in = [null];
        }
        // NOT IN () is always true
        if (cond[field] && cond[field].$nin && cond[field].$nin.length === 0) {
            delete cond[field];
        }
    }
    getLockSQL(lockMode) {
        if (lockMode === unit_of_work_1.LockMode.PESSIMISTIC_READ) {
            return ' ' + this.platform.getReadLockSQL();
        }
        if (lockMode === unit_of_work_1.LockMode.PESSIMISTIC_WRITE) {
            return ' ' + this.platform.getWriteLockSQL();
        }
        if (lockMode === unit_of_work_1.LockMode.OPTIMISTIC && this.metadata[this.entityName] && !this.metadata[this.entityName].versionProperty) {
            throw utils_1.ValidationError.lockFailed(this.entityName);
        }
        return '';
    }
    updateVersionProperty(set) {
        const meta = this.metadata[this.entityName];
        if (!meta || !meta.versionProperty) {
            return;
        }
        const versionProperty = meta.properties[meta.versionProperty];
        let sql = `${this.wrap(versionProperty.fieldName)} = `;
        if (versionProperty.type.toLowerCase() === 'date') {
            sql += this.platform.getCurrentTimestampSQL(versionProperty.length);
        }
        else {
            sql += this.wrap(versionProperty.fieldName) + ' + 1';
        }
        set.push(sql);
    }
    processComplexParam(key, cond) {
        // unwind parameters when ? found in field name
        const customExpression = key.match(/\(.*\)| |\?/) && Array.isArray(cond);
        if (customExpression) {
            const count = key.concat('?').match(/\?/g).length - 1;
            return cond.slice(0, count).map((c) => utils_1.Utils.isObject(c) ? JSON.stringify(c) : c).concat(cond.slice(count));
        }
        const operator = Object.keys(QueryBuilderHelper.OPERATORS).find(op => op in cond);
        if (operator) {
            return utils_1.Utils.asArray(cond[operator]);
        }
        return utils_1.Utils.asArray(cond);
    }
    prefixAndWrap(field) {
        if (!this.isPrefixed(field)) {
            return this.wrap(this.fieldName(field, this.alias));
        }
        const [a, f] = field.split('.');
        return this.wrap(a) + '.' + this.wrap(this.fieldName(f, a));
    }
    getGroupWhereParams(key, cond) {
        if (key === '$and' || key === '$or') {
            return utils_1.Utils.flatten(cond.map((sub) => this.getWhereParams(sub)));
        }
        return this.getWhereParams(cond);
    }
    processValue(field, value) {
        if (value === null) {
            return ' IS NULL';
        }
        if (value instanceof RegExp) {
            return ' LIKE ?';
        }
        if (utils_1.Utils.isObject(value) && !(value instanceof Date)) {
            return this.processObjectValue(value);
        }
        const wildcards = field.concat('?').match(/\?/g).length - 1;
        if (Array.isArray(value) && value.length === wildcards) {
            return '';
        }
        return ' = ?';
    }
    processObjectValue(value) {
        let ret = '';
        for (const [op, replacement] of Object.entries(QueryBuilderHelper.OPERATORS)) {
            if (!(op in value)) {
                continue;
            }
            const token = Array.isArray(value[op]) ? `(${value[op].map(() => '?').join(', ')})` : '?';
            ret = ` ${replacement} ${token}`;
            break;
        }
        return ret;
    }
    getGroupQueryCondition(type, operator, subCondition) {
        const glue = QueryBuilderHelper.GROUP_OPERATORS[operator];
        const group = subCondition.map(sub => {
            const cond = this.getQueryCondition(type, sub);
            return cond.length > 1 ? '(' + cond.join(` AND `) + ')' : cond[0];
        });
        return '(' + group.join(` ${glue} `) + ')';
    }
    isPrefixed(field) {
        return new RegExp(`${this.quoteChar}?\\w+${this.quoteChar}?\\.`).test(field);
    }
    fieldName(field, alias) {
        const entityName = this.aliasMap[alias] || this.entityName;
        const prop = this.metadata[entityName] ? this.metadata[entityName].properties[field] : false;
        return prop ? prop.fieldName : field;
    }
}
QueryBuilderHelper.GROUP_OPERATORS = {
    $and: 'AND',
    $or: 'OR',
};
QueryBuilderHelper.OPERATORS = {
    $eq: '=',
    $in: 'IN',
    $nin: 'NOT IN',
    $gt: '>',
    $gte: '>=',
    $lt: '<',
    $lte: '<=',
    $ne: '!=',
};
exports.QueryBuilderHelper = QueryBuilderHelper;
