"use strict";
Object.defineProperty(exports, "__esModule", { value: true });
const __1 = require("..");
class SchemaGenerator {
    constructor(driver, metadata) {
        this.driver = driver;
        this.metadata = metadata;
        this.platform = this.driver.getPlatform();
        this.helper = this.platform.getSchemaHelper();
    }
    generate() {
        let ret = this.helper.getSchemaBeginning();
        Object.values(this.metadata).forEach(meta => {
            ret += this.helper.dropTable(meta) + '\n';
            ret += this.createTable(meta) + '\n';
        });
        Object.values(this.metadata).forEach(meta => {
            ret += this.createForeignKeys(meta);
        });
        ret += this.helper.getSchemaEnd();
        return ret;
    }
    createTable(meta) {
        const pkProp = meta.properties[meta.primaryKey];
        let ret = '';
        if (this.helper.supportsSequences() && pkProp.type === 'number') {
            ret += `CREATE SEQUENCE ${this.helper.quoteIdentifier(meta.collection + '_seq')};\n`;
        }
        ret += `CREATE TABLE ${this.helper.quoteIdentifier(meta.collection)} (\n`;
        Object
            .values(meta.properties)
            .filter(prop => this.shouldHaveColumn(prop))
            .forEach(prop => ret += '  ' + this.createTableColumn(meta, prop) + ',\n');
        if (this.helper.supportsSchemaConstraints()) {
            ret += this.createIndexes(meta);
        }
        else {
            ret = ret.substr(0, ret.length - 2) + '\n';
        }
        ret += `)${this.helper.getSchemaTableEnd()};\n\n`;
        return ret;
    }
    shouldHaveColumn(prop) {
        if (prop.persist === false) {
            return false;
        }
        if (prop.reference === __1.ReferenceType.SCALAR) {
            return true;
        }
        if (!this.helper.supportsSchemaConstraints()) {
            return false;
        }
        return prop.reference === __1.ReferenceType.MANY_TO_ONE || (prop.reference === __1.ReferenceType.ONE_TO_ONE && prop.owner);
    }
    createTableColumn(meta, prop, alter = false) {
        const fieldName = prop.fieldName;
        const ret = this.helper.quoteIdentifier(fieldName) + ' ' + this.type(prop);
        const nullable = (alter && this.platform.requiresNullableForAlteringColumn()) || prop.nullable;
        if (prop.primary) {
            return ret + this.helper.createPrimaryKeyColumn(meta, prop);
        }
        return ret + this.helper.createColumn(meta, prop, nullable);
    }
    createIndexes(meta) {
        let ret = `  PRIMARY KEY (${this.helper.quoteIdentifier(meta.properties[meta.primaryKey].fieldName)})`;
        if (this.helper.indexForeignKeys()) {
            Object
                .values(meta.properties)
                .filter(prop => prop.reference === __1.ReferenceType.MANY_TO_ONE)
                .forEach(prop => ret += `,\n  KEY ${this.helper.quoteIdentifier(prop.fieldName)} (${this.helper.quoteIdentifier(prop.fieldName)})`);
        }
        return ret + '\n';
    }
    createForeignKeys(meta) {
        const ret = `ALTER TABLE ${this.helper.quoteIdentifier(meta.collection)}`;
        let i = 1;
        const constraints = Object
            .values(meta.properties)
            .filter(prop => prop.reference === __1.ReferenceType.MANY_TO_ONE || (prop.reference === __1.ReferenceType.ONE_TO_ONE && prop.owner))
            .map(prop => this.createForeignKey(meta, prop, i++));
        if (constraints.length === 0) {
            return '';
        }
        if (this.helper.supportsSchemaMultiAlter()) {
            return ret + '\n ' + constraints.join(',\n ') + ';\n\n\n';
        }
        return constraints.map(c => ret + c + ';').join('\n') + '\n\n';
    }
    createForeignKey(meta, prop, index) {
        if (this.helper.supportsSchemaConstraints()) {
            return this.createForeignConstraint(meta, prop, index);
        }
        let ret = ' ADD ' + this.createTableColumn(meta, prop, true) + ' ';
        ret += this.createForeignKeyReference(prop);
        return ret;
    }
    createForeignConstraint(meta, prop, index) {
        let ret = ' ADD CONSTRAINT ' + this.helper.quoteIdentifier(meta.collection + '_ibfk_' + index);
        ret += ` FOREIGN KEY (${this.helper.quoteIdentifier(prop.fieldName)}) `;
        ret += this.createForeignKeyReference(prop);
        return ret;
    }
    createForeignKeyReference(prop) {
        const meta2 = this.metadata[prop.type];
        const pk2 = meta2.properties[meta2.primaryKey].fieldName;
        let ret = `REFERENCES ${this.helper.quoteIdentifier(meta2.collection)} (${this.helper.quoteIdentifier(pk2)})`;
        const cascade = prop.cascade.includes(__1.Cascade.REMOVE) || prop.cascade.includes(__1.Cascade.ALL);
        ret += ` ON DELETE ${cascade ? 'CASCADE' : 'SET NULL'}`;
        if (prop.cascade.includes(__1.Cascade.PERSIST) || prop.cascade.includes(__1.Cascade.ALL)) {
            ret += ' ON UPDATE CASCADE';
        }
        return ret;
    }
    type(prop, foreignKey) {
        const type = this.helper.getTypeDefinition(prop);
        if (prop.reference !== __1.ReferenceType.SCALAR) {
            const meta = this.metadata[prop.type];
            return this.type(meta.properties[meta.primaryKey], prop);
        }
        if (prop.type === 'number' && prop.primary) {
            return type + this.helper.getUnsignedSuffix(foreignKey || prop);
        }
        return type;
    }
}
exports.SchemaGenerator = SchemaGenerator;
