"use strict";
Object.defineProperty(exports, "__esModule", { value: true });
const metadata_1 = require("../metadata");
const entity_1 = require("../entity");
const ChangeSetComputer_1 = require("./ChangeSetComputer");
const ChangeSetPersister_1 = require("./ChangeSetPersister");
const ChangeSet_1 = require("./ChangeSet");
const utils_1 = require("../utils");
const __1 = require("..");
class UnitOfWork {
    constructor(em) {
        this.em = em;
        /** map of references to managed entities */
        this.identityMap = {};
        /** holds copy of identity map so we can compute changes when persisting managed entities */
        this.originalEntityData = {};
        /** map of wrapped primary keys so we can compute change set without eager commit */
        this.identifierMap = {};
        this.persistStack = [];
        this.removeStack = [];
        this.orphanRemoveStack = [];
        this.changeSets = [];
        this.extraUpdates = [];
        this.metadata = metadata_1.MetadataStorage.getMetadata();
        this.platform = this.em.getDriver().getPlatform();
        this.changeSetComputer = new ChangeSetComputer_1.ChangeSetComputer(this.em.getValidator(), this.originalEntityData, this.identifierMap);
        this.changeSetPersister = new ChangeSetPersister_1.ChangeSetPersister(this.em.getDriver(), this.identifierMap);
    }
    merge(entity, visited = [], mergeData = true) {
        if (!entity.__primaryKey) {
            return;
        }
        this.identityMap[`${entity.constructor.name}-${entity.__serializedPrimaryKey}`] = entity;
        if (!this.originalEntityData[entity.__uuid] || mergeData) {
            this.originalEntityData[entity.__uuid] = utils_1.Utils.prepareEntity(entity);
        }
        this.cascade(entity, entity_1.Cascade.MERGE, visited);
    }
    getById(entityName, id) {
        const token = `${entityName}-${this.platform.normalizePrimaryKey(id)}`;
        return this.identityMap[token];
    }
    tryGetById(entityName, where) {
        const pk = utils_1.Utils.extractPK(where, this.metadata[entityName]);
        if (!pk) {
            return null;
        }
        return this.getById(entityName, pk);
    }
    getIdentityMap() {
        return this.identityMap;
    }
    persist(entity, visited = []) {
        if (this.persistStack.includes(entity)) {
            return;
        }
        if (!entity.__primaryKey) {
            this.identifierMap[entity.__uuid] = new entity_1.EntityIdentifier();
        }
        this.persistStack.push(entity);
        this.cleanUpStack(this.removeStack, entity);
        this.cascade(entity, entity_1.Cascade.PERSIST, visited);
    }
    remove(entity, visited = []) {
        if (this.removeStack.includes(entity)) {
            return;
        }
        if (entity.__primaryKey) {
            this.removeStack.push(entity);
        }
        this.cleanUpStack(this.persistStack, entity);
        this.unsetIdentity(entity);
        this.cascade(entity, entity_1.Cascade.REMOVE, visited);
    }
    async commit() {
        this.computeChangeSets();
        if (this.changeSets.length === 0) {
            return this.postCommitCleanup(); // nothing to do, do not start transaction
        }
        const driver = this.em.getDriver();
        const runInTransaction = !driver.isInTransaction() && driver.getPlatform().supportsTransactions();
        const promise = utils_1.Utils.runSerial(this.changeSets, changeSet => this.commitChangeSet(changeSet));
        if (runInTransaction) {
            await driver.transactional(() => promise);
        }
        else {
            await promise;
        }
        this.postCommitCleanup();
    }
    async lock(entity, mode, version) {
        if (!this.getById(entity.constructor.name, entity.__primaryKey)) {
            throw utils_1.ValidationError.entityNotManaged(entity);
        }
        const meta = this.metadata[entity.constructor.name];
        if (mode === __1.LockMode.OPTIMISTIC) {
            await this.lockOptimistic(entity, meta, version);
        }
        else if ([__1.LockMode.NONE, __1.LockMode.PESSIMISTIC_READ, __1.LockMode.PESSIMISTIC_WRITE].includes(mode)) {
            await this.lockPessimistic(entity, mode);
        }
    }
    clear() {
        Object.keys(this.identityMap).forEach(key => delete this.identityMap[key]);
        Object.keys(this.originalEntityData).forEach(key => delete this.originalEntityData[key]);
        this.postCommitCleanup();
    }
    unsetIdentity(entity) {
        delete this.identityMap[`${entity.constructor.name}-${entity.__serializedPrimaryKey}`];
        delete this.identifierMap[entity.__uuid];
        delete this.originalEntityData[entity.__uuid];
    }
    computeChangeSets() {
        this.changeSets.length = 0;
        Object.values(this.identityMap)
            .filter(entity => !this.removeStack.includes(entity) && !this.orphanRemoveStack.includes(entity))
            .forEach(entity => this.persist(entity));
        while (this.persistStack.length) {
            this.findNewEntities(this.persistStack.shift());
        }
        while (this.extraUpdates.length) {
            const extraUpdate = this.extraUpdates.shift();
            extraUpdate[0][extraUpdate[1]] = extraUpdate[2];
            const changeSet = this.changeSetComputer.computeChangeSet(extraUpdate[0]);
            this.changeSets.push(changeSet);
        }
        for (const entity of Object.values(this.orphanRemoveStack)) {
            this.remove(entity);
        }
        for (const entity of Object.values(this.removeStack)) {
            const meta = this.metadata[entity.constructor.name];
            this.changeSets.push({ entity, type: ChangeSet_1.ChangeSetType.DELETE, name: meta.name, collection: meta.collection, payload: {} });
        }
    }
    scheduleOrphanRemoval(entity) {
        this.orphanRemoveStack.push(entity);
    }
    cancelOrphanRemoval(entity) {
        this.cleanUpStack(this.orphanRemoveStack, entity);
    }
    findNewEntities(entity, visited = []) {
        visited.push(entity);
        const meta = this.metadata[entity.constructor.name];
        if (!entity.__primaryKey && !this.identifierMap[entity.__uuid]) {
            this.identifierMap[entity.__uuid] = new entity_1.EntityIdentifier();
        }
        for (const prop of Object.values(meta.properties)) {
            const reference = entity[prop.name];
            this.processReference(entity, prop, reference, visited);
        }
        const changeSet = this.changeSetComputer.computeChangeSet(entity);
        if (changeSet) {
            this.changeSets.push(changeSet);
            this.cleanUpStack(this.persistStack, entity);
            this.originalEntityData[entity.__uuid] = utils_1.Utils.prepareEntity(entity);
        }
    }
    processReference(parent, prop, reference, visited) {
        const isToOne = prop.reference === entity_1.ReferenceType.MANY_TO_ONE || prop.reference === entity_1.ReferenceType.ONE_TO_ONE;
        if (isToOne && reference) {
            return this.processToOneReference(parent, prop, reference, visited);
        }
        if (utils_1.Utils.isCollection(reference, prop, entity_1.ReferenceType.MANY_TO_MANY) && reference.isDirty()) {
            this.processToManyReference(reference, visited, parent, prop);
        }
    }
    processToOneReference(parent, prop, reference, visited) {
        if (!this.hasIdentifier(reference) && visited.includes(reference)) {
            this.extraUpdates.push([parent, prop.name, reference]);
            delete parent[prop.name];
        }
        if (!this.originalEntityData[reference.__uuid]) {
            this.findNewEntities(reference, visited);
        }
    }
    processToManyReference(reference, visited, parent, prop) {
        if (this.isCollectionSelfReferenced(reference, visited)) {
            this.extraUpdates.push([parent, prop.name, reference]);
            parent[prop.name] = new entity_1.Collection(parent);
            return;
        }
        reference.getItems()
            .filter(item => !this.originalEntityData[item.__uuid])
            .forEach(item => this.findNewEntities(item, visited));
    }
    async commitChangeSet(changeSet) {
        const type = changeSet.type.charAt(0).toUpperCase() + changeSet.type.slice(1);
        await this.runHooks(`before${type}`, changeSet.entity, changeSet.payload);
        await this.changeSetPersister.persistToDatabase(changeSet);
        switch (changeSet.type) {
            case ChangeSet_1.ChangeSetType.CREATE:
                this.em.merge(changeSet.entity);
                break;
            case ChangeSet_1.ChangeSetType.UPDATE:
                this.merge(changeSet.entity);
                break;
            case ChangeSet_1.ChangeSetType.DELETE:
                this.unsetIdentity(changeSet.entity);
                break;
        }
        await this.runHooks(`after${type}`, changeSet.entity);
    }
    async runHooks(type, entity, payload) {
        const hooks = this.metadata[entity.constructor.name].hooks;
        if (hooks && hooks[type] && hooks[type].length > 0) {
            const copy = utils_1.Utils.copy(entity);
            await utils_1.Utils.runSerial(hooks[type], hook => entity[hook]());
            if (payload) {
                Object.assign(payload, utils_1.Utils.diffEntities(copy, entity));
            }
        }
    }
    /**
     * clean up persist/remove stack from previous persist/remove calls for this entity done before flushing
     */
    cleanUpStack(stack, entity) {
        for (const index in stack) {
            if (stack[index] === entity) {
                stack.splice(+index, 1);
            }
        }
    }
    postCommitCleanup() {
        Object.keys(this.identifierMap).forEach(key => delete this.identifierMap[key]);
        this.persistStack.length = 0;
        this.removeStack.length = 0;
        this.orphanRemoveStack.length = 0;
        this.changeSets.length = 0;
    }
    hasIdentifier(entity) {
        const pk = this.metadata[entity.constructor.name].primaryKey;
        if (entity[pk]) {
            return true;
        }
        return this.identifierMap[entity.__uuid] && this.identifierMap[entity.__uuid].getValue();
    }
    cascade(entity, type, visited) {
        if (visited.includes(entity)) {
            return;
        }
        visited.push(entity);
        switch (type) {
            case entity_1.Cascade.PERSIST:
                this.persist(entity, visited);
                break;
            case entity_1.Cascade.MERGE:
                this.merge(entity, visited, false);
                break;
            case entity_1.Cascade.REMOVE:
                this.remove(entity, visited);
                break;
        }
        const meta = this.metadata[entity.constructor.name];
        for (const prop of Object.values(meta.properties)) {
            this.cascadeReference(entity, prop, type, visited);
        }
    }
    cascadeReference(entity, prop, type, visited) {
        this.fixMissingReference(entity, prop);
        if (!this.shouldCascade(prop, type)) {
            return;
        }
        if ((prop.reference === entity_1.ReferenceType.MANY_TO_ONE || prop.reference === entity_1.ReferenceType.ONE_TO_ONE) && entity[prop.name]) {
            return this.cascade(entity[prop.name], type, visited);
        }
        const collection = entity[prop.name];
        const requireFullyInitialized = type === entity_1.Cascade.PERSIST; // only cascade persist needs fully initialized items
        if ([entity_1.ReferenceType.ONE_TO_MANY, entity_1.ReferenceType.MANY_TO_MANY].includes(prop.reference) && collection.isInitialized(requireFullyInitialized)) {
            collection.getItems().forEach(item => this.cascade(item, type, visited));
        }
    }
    isCollectionSelfReferenced(collection, visited) {
        const filtered = collection.getItems().filter(item => !this.originalEntityData[item.__uuid]);
        return filtered.some(items => visited.includes(items));
    }
    shouldCascade(prop, type) {
        if (type === entity_1.Cascade.REMOVE && prop.orphanRemoval) {
            return true;
        }
        return prop.cascade && (prop.cascade.includes(type) || prop.cascade.includes(entity_1.Cascade.ALL));
    }
    async lockPessimistic(entity, mode) {
        if (!this.em.getDriver().isInTransaction()) {
            throw utils_1.ValidationError.transactionRequired();
        }
        const qb = this.em.createQueryBuilder(entity.constructor.name);
        await qb.select('1').where({ [entity.__primaryKeyField]: entity.__primaryKey }).setLockMode(mode).execute();
    }
    async lockOptimistic(entity, meta, version) {
        if (!meta.versionProperty) {
            throw utils_1.ValidationError.notVersioned(meta);
        }
        if (!utils_1.Utils.isDefined(version)) {
            return;
        }
        if (!entity.isInitialized()) {
            await entity.init();
        }
        if (entity[meta.versionProperty] !== version) {
            throw utils_1.ValidationError.lockFailedVersionMismatch(entity, version, entity[meta.versionProperty]);
        }
    }
    fixMissingReference(entity, prop) {
        const reference = entity[prop.name];
        if ([entity_1.ReferenceType.MANY_TO_ONE, entity_1.ReferenceType.ONE_TO_ONE].includes(prop.reference) && reference && !utils_1.Utils.isEntity(reference)) {
            entity[prop.name] = this.em.getReference(prop.type, reference);
        }
        const isCollection = [entity_1.ReferenceType.ONE_TO_MANY, entity_1.ReferenceType.MANY_TO_MANY].includes(prop.reference);
        if (isCollection && Array.isArray(reference)) {
            entity[prop.name] = new entity_1.Collection(entity);
            entity[prop.name].set(reference);
            entity[prop.name].setDirty();
        }
    }
}
exports.UnitOfWork = UnitOfWork;
