"use strict";
var __importDefault = (this && this.__importDefault) || function (mod) {
    return (mod && mod.__esModule) ? mod : { "default": mod };
};
Object.defineProperty(exports, "__esModule", { value: true });
const fast_deep_equal_1 = __importDefault(require("fast-deep-equal"));
const clone_1 = __importDefault(require("clone"));
const metadata_1 = require("../metadata");
const entity_1 = require("../entity");
class Utils {
    static isDefined(data) {
        return typeof data !== 'undefined';
    }
    static isObject(o) {
        return !!o && typeof o === 'object' && !Array.isArray(o);
    }
    static isString(s) {
        return typeof s === 'string';
    }
    static isNumber(s) {
        return typeof s === 'number';
    }
    static equals(a, b) {
        return fast_deep_equal_1.default(a, b);
    }
    static unique(items) {
        return [...new Set(items)];
    }
    static flatten(arrays) {
        return [].concat(...arrays);
    }
    static merge(target, ...sources) {
        if (!sources.length) {
            return target;
        }
        const source = sources.shift();
        if (Utils.isObject(target) && Utils.isObject(source)) {
            Object.entries(source).forEach(([key, value]) => {
                if (Utils.isObject(value)) {
                    if (!target[key]) {
                        Object.assign(target, { [key]: {} });
                    }
                    Utils.merge(target[key], value);
                }
                else {
                    Object.assign(target, { [key]: value });
                }
            });
        }
        return Utils.merge(target, ...sources);
    }
    static diff(a, b) {
        const ret = {};
        Object.keys(b).forEach(k => {
            if (Utils.equals(a[k], b[k])) {
                return;
            }
            ret[k] = b[k];
        });
        return ret;
    }
    static diffEntities(a, b) {
        return Utils.diff(Utils.prepareEntity(a), Utils.prepareEntity(b));
    }
    static prepareEntity(e) {
        const metadata = metadata_1.MetadataStorage.getMetadata();
        const meta = metadata[e.constructor.name];
        const ret = Utils.copy(e);
        delete ret.__initialized;
        // remove collections and references
        Object.values(meta.properties).forEach(prop => {
            const pk = () => metadata[prop.type].primaryKey;
            const name = prop.name;
            if (e[name] instanceof entity_1.ArrayCollection || (Utils.isEntity(e[name]) && !e[name][pk()])) {
                return delete ret[name];
            }
            if (Utils.isEntity(e[name])) {
                return ret[prop.name] = ret[prop.name][pk()];
            }
        });
        // remove unknown properties
        Object.keys(e).forEach(prop => {
            if (!meta.properties[prop] || meta.properties[prop].persist === false) {
                delete ret[prop];
            }
        });
        return ret;
    }
    static copy(entity) {
        return clone_1.default(entity);
    }
    static asArray(data) {
        if (typeof data === 'undefined') {
            return [];
        }
        return Array.isArray(data) ? data : [data];
    }
    static renameKey(payload, from, to) {
        if (Utils.isObject(payload) && from in payload && !(to in payload)) {
            payload[to] = payload[from];
            delete payload[from];
        }
    }
    static getParamNames(func) {
        const STRIP_COMMENTS = /((\/\/.*$)|(\/\*[\s\S]*?\*\/))/mg;
        const ARGUMENT_NAMES = /([^\s,]+)/g;
        const fnStr = func.toString().replace(STRIP_COMMENTS, ''); // strip comments
        let paramsStr = fnStr.slice(fnStr.indexOf('(') + 1, fnStr.indexOf(')')); // extract params
        paramsStr = paramsStr.replace(/{[^}]+}/g, '{}'); // simplify object default values like `a = { ... }`
        paramsStr = paramsStr.replace(/\[[^\]]+]/g, '[]'); // simplify array default values like `a = [ ... ]`
        const result = paramsStr.match(ARGUMENT_NAMES);
        if (result === null) {
            return [];
        }
        // handle class with no constructor
        if (result.length > 0 && result[0] === 'class') {
            return [];
        }
        // strip default values
        for (let i = 0; i < result.length; i++) {
            if (result[i].includes('=')) {
                result[i] = result[i].split('=')[0];
                result.splice(i + 1, 1);
            }
        }
        return result.filter(i => i); // filter out empty strings
    }
    static isPrimaryKey(key) {
        return Utils.isString(key) || typeof key === 'number' || Utils.isObjectID(key);
    }
    static extractPK(data, meta) {
        if (Utils.isPrimaryKey(data)) {
            return data;
        }
        if (Utils.isObject(data) && meta) {
            return data[meta.primaryKey] || data[meta.serializedPrimaryKey] || null;
        }
        return null;
    }
    static isEntity(data) {
        return Utils.isObject(data) && !!data.__entity;
    }
    static isObjectID(key) {
        return Utils.isObject(key) && key.constructor.name === 'ObjectID';
    }
    static isEmpty(data) {
        if (Array.isArray(data)) {
            return data.length === 0;
        }
        if (Utils.isObject(data)) {
            return Object.keys(data).length === 0;
        }
        return !data;
    }
    static className(classOrName) {
        if (Utils.isString(classOrName)) {
            return classOrName;
        }
        return classOrName.name;
    }
    /**
     * uses some dark magic to get source path to caller where decorator is used
     */
    static lookupPathFromDecorator(meta, stack) {
        if (meta.path) {
            return meta.path;
        }
        // use some dark magic to get source path to caller
        stack = stack || new Error().stack.split('\n');
        let line = stack.findIndex(line => line.includes('__decorate'));
        if (line === -1) {
            return meta.path;
        }
        if (stack[line].includes('node_modules/tslib/tslib')) {
            line++;
        }
        meta.path = Utils.normalizePath(stack[line].match(/\((.*):\d+:\d+\)/)[1]);
        return meta.path;
    }
    static getObjectType(value) {
        const objectType = Object.prototype.toString.call(value);
        return objectType.match(/\[object (\w+)]/)[1].toLowerCase();
    }
    static async runSerial(items, cb) {
        const ret = [];
        for (const item of items) {
            ret.push(await cb(item));
        }
        return ret;
    }
    static isCollection(item, prop, type) {
        if (!(item instanceof entity_1.Collection)) {
            return false;
        }
        return !(prop && type) || prop.reference === type;
    }
    static normalizePath(...parts) {
        return parts.join('/').replace(/\\/g, '/');
    }
}
exports.Utils = Utils;
